@@ -2,43 +2,60 @@
 Renders the command line on the console.
 (Redraws parts of the input line that were changed.)
 """
-from __future__ import unicode_literals
-
-import threading
-import time
+from asyncio import FIRST_COMPLETED, Future, sleep, wait
 from collections import deque
-
-from six.moves import range
-
-from prompt_toolkit.eventloop import (
-    From,
-    Future,
-    ensure_future,
-    get_event_loop,
+from enum import Enum
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Deque,
+    Dict,
+    Hashable,
+    Optional,
+    Tuple,
 )
-from prompt_toolkit.filters import to_filter
-from prompt_toolkit.formatted_text import to_formatted_text
+
+from prompt_toolkit.application.current import get_app
+from prompt_toolkit.data_structures import Point, Size
+from prompt_toolkit.filters import FilterOrBool, to_filter
+from prompt_toolkit.formatted_text import AnyFormattedText, to_formatted_text
 from prompt_toolkit.input.base import Input
 from prompt_toolkit.layout.mouse_handlers import MouseHandlers
-from prompt_toolkit.layout.screen import Point, Screen, WritePosition
+from prompt_toolkit.layout.screen import Char, Screen, WritePosition
 from prompt_toolkit.output import ColorDepth, Output
 from prompt_toolkit.styles import (
+    Attrs,
     BaseStyle,
     DummyStyleTransformation,
     StyleTransformation,
 )
 from prompt_toolkit.utils import is_windows
 
+if TYPE_CHECKING:
+    from prompt_toolkit.application import Application
+    from prompt_toolkit.layout.layout import Layout
+
+
 __all__ = [
     'Renderer',
     'print_formatted_text',
 ]
 
 
-def _output_screen_diff(app, output, screen, current_pos, color_depth,
-                        previous_screen=None, last_style=None, is_done=False,
-                        full_screen=False, attrs_for_style_string=None,
-                        size=None, previous_width=0):  # XXX: drop is_done
+def _output_screen_diff(
+        app: 'Application[Any]',
+        output: Output,
+        screen: Screen,
+        current_pos: Point,
+        color_depth: ColorDepth,
+        previous_screen: Optional[Screen],
+        last_style: Optional[str],
+        is_done: bool,  # XXX: drop is_done
+        full_screen: bool,
+        attrs_for_style_string: '_StyleStringToAttrsCache',
+        size: Size,
+        previous_width: int) -> Tuple[Point, Optional[str]]:
     """
     Render the diff between this screen and the previous screen.
 
@@ -60,9 +77,6 @@ def _output_screen_diff(app, output, screen, current_pos, color_depth,
     """
     width, height = size.columns, size.rows
 
-    #: Remember the last printed character.
-    last_style = [last_style]  # nonlocal
-
     #: Variable for capturing the output.
     write = output.write
     write_raw = output.write_raw
@@ -78,12 +92,13 @@ def _output_screen_diff(app, output, screen, current_pos, color_depth,
     # Hide cursor before rendering. (Avoid flickering.)
     output.hide_cursor()
 
-    def reset_attributes():
+    def reset_attributes() -> None:
         " Wrapper around Output.reset_attributes. "
+        nonlocal last_style
         _output_reset_attributes()
-        last_style[0] = None  # Forget last char after resetting attributes.
+        last_style = None  # Forget last char after resetting attributes.
 
-    def move_cursor(new):
+    def move_cursor(new: Point) -> Point:
         " Move cursor to this `new` point. Returns the given Point. "
         current_x, current_y = current_pos.x, current_pos.y
 
@@ -110,15 +125,15 @@ def _output_screen_diff(app, output, screen, current_pos, color_depth,
 
         return new
 
-    def output_char(char):
+    def output_char(char: Char) -> None:
         """
         Write the output of this character.
         """
+        nonlocal last_style
+
         # If the last printed character has the same style, don't output the
         # style again.
-        the_last_style = last_style[0]  # Either `None` or a style string.
-
-        if the_last_style == char.style:
+        if last_style == char.style:
             write(char.char)
         else:
             # Look up `Attr` for this style string. Only set attributes if different.
@@ -126,11 +141,11 @@ def _output_screen_diff(app, output, screen, current_pos, color_depth,
             # Note that an empty style string can have formatting that needs to
             # be applied, because of style transformations.
             new_attrs = attrs_for_style_string[char.style]
-            if not the_last_style or new_attrs != attrs_for_style_string[the_last_style]:
+            if not last_style or new_attrs != attrs_for_style_string[last_style]:
                 _output_set_attributes(new_attrs, color_depth)
 
             write(char.char)
-            last_style[0] = char.style
+            last_style = char.style
 
     # Render for the first time: reset styling.
     if not previous_screen:
@@ -230,26 +245,25 @@ def _output_screen_diff(app, output, screen, current_pos, color_depth,
     if screen.show_cursor or is_done:
         output.show_cursor()
 
-    return current_pos, last_style[0]
+    return current_pos, last_style
 
 
 class HeightIsUnknownError(Exception):
     " Information unavailable. Did not yet receive the CPR response. "
 
 
-class _StyleStringToAttrsCache(dict):
+class _StyleStringToAttrsCache(Dict[str, Attrs]):
     """
     A cache structure that maps style strings to :class:`.Attr`.
     (This is an important speed up.)
     """
-    def __init__(self, get_attrs_for_style_str, style_transformation):
-        assert callable(get_attrs_for_style_str)
-        assert isinstance(style_transformation, StyleTransformation)
+    def __init__(self, get_attrs_for_style_str: Callable[['str'], Attrs],
+                 style_transformation: StyleTransformation) -> None:
 
         self.get_attrs_for_style_str = get_attrs_for_style_str
         self.style_transformation = style_transformation
 
-    def __missing__(self, style_str):
+    def __missing__(self, style_str: str) -> Attrs:
         attrs = self.get_attrs_for_style_str(style_str)
         attrs = self.style_transformation.transform_attrs(attrs)
 
@@ -257,14 +271,14 @@ class _StyleStringToAttrsCache(dict):
         return attrs
 
 
-class CPR_Support(object):
+class CPR_Support(Enum):
     " Enum: whether or not CPR is supported. "
     SUPPORTED = 'SUPPORTED'
     NOT_SUPPORTED = 'NOT_SUPPORTED'
     UNKNOWN = 'UNKNOWN'
 
 
-class Renderer(object):
+class Renderer:
     """
     Typical usage:
 
@@ -276,13 +290,14 @@ class Renderer(object):
     """
     CPR_TIMEOUT = 2  # Time to wait until we consider CPR to be not supported.
 
-    def __init__(self, style, output, input, full_screen=False,
-                 mouse_support=False, cpr_not_supported_callback=None):
-
-        assert isinstance(style, BaseStyle)
-        assert isinstance(output, Output)
-        assert isinstance(input, Input)
-        assert callable(cpr_not_supported_callback) or cpr_not_supported_callback is None
+    def __init__(
+            self,
+            style: BaseStyle,
+            output: Output,
+            input: Input,
+            full_screen: bool = False,
+            mouse_support: FilterOrBool = False,
+            cpr_not_supported_callback: Optional[Callable[[], None]] = None) -> None:
 
         self.style = style
         self.output = output
@@ -296,20 +311,22 @@ class Renderer(object):
         self._bracketed_paste_enabled = False
 
         # Future set when we are waiting for a CPR flag.
-        self._waiting_for_cpr_futures = deque()
+        self._waiting_for_cpr_futures: Deque[Future[None]] = deque()
         self.cpr_support = CPR_Support.UNKNOWN
         if not input.responds_to_cpr:
             self.cpr_support = CPR_Support.NOT_SUPPORTED
 
         # Cache for the style.
-        self._attrs_for_style = None
-        self._last_style_hash = None
-        self._last_transformation_hash = None
-        self._last_color_depth = None
+        self._attrs_for_style: Optional[_StyleStringToAttrsCache] = None
+        self._last_style_hash: Optional[Hashable] = None
+        self._last_transformation_hash: Optional[Hashable] = None
+        self._last_color_depth: Optional[ColorDepth] = None
 
         self.reset(_scroll=True)
 
-    def reset(self, _scroll=False, leave_alternate_screen=True):
+    def reset(self, _scroll: bool = False,
+              leave_alternate_screen: bool = True) -> None:
+
         # Reset position
         self._cursor_pos = Point(x=0, y=0)
 
@@ -317,9 +334,9 @@ class Renderer(object):
         # we can create a `diff` between two screens and only output the
         # difference. It's also to remember the last height. (To show for
         # instance a toolbar at the bottom position.)
-        self._last_screen = None
-        self._last_size = None
-        self._last_style = None
+        self._last_screen: Optional[Screen] = None
+        self._last_size: Optional[Size] = None
+        self._last_style: Optional[str] = None
 
         # Default MouseHandlers. (Just empty.)
         self.mouse_handlers = MouseHandlers()
@@ -352,7 +369,7 @@ class Renderer(object):
         self.output.flush()
 
     @property
-    def last_rendered_screen(self):
+    def last_rendered_screen(self) -> Optional[Screen]:
         """
         The `Screen` class that was generated during the last rendering.
         This can be `None`.
@@ -360,7 +377,7 @@ class Renderer(object):
         return self._last_screen
 
     @property
-    def height_is_known(self):
+    def height_is_known(self) -> bool:
         """
         True when the height from the cursor until the bottom of the terminal
         is known. (It's often nicer to draw bottom toolbars only if the height
@@ -370,7 +387,7 @@ class Renderer(object):
             is_windows()  # On Windows, we don't have to wait for a CPR.
 
     @property
-    def rows_above_layout(self):
+    def rows_above_layout(self) -> int:
         """
         Return the number of rows visible in the terminal above the layout.
         """
@@ -383,7 +400,7 @@ class Renderer(object):
         else:
             raise HeightIsUnknownError('Rows above layout is unknown.')
 
-    def request_absolute_cursor_position(self):
+    def request_absolute_cursor_position(self) -> None:
         """
         Get current cursor position.
 
@@ -412,7 +429,7 @@ class Renderer(object):
             if self.cpr_support == CPR_Support.NOT_SUPPORTED:
                 return
 
-            def do_cpr():
+            def do_cpr() -> None:
                 # Asks for a cursor position report (CPR).
                 self._waiting_for_cpr_futures.append(Future())
                 self.output.ask_for_cpr()
@@ -425,8 +442,8 @@ class Renderer(object):
             elif self.cpr_support == CPR_Support.UNKNOWN and not self.waiting_for_cpr:
                 do_cpr()
 
-                def timer():
-                    time.sleep(self.CPR_TIMEOUT)
+                async def timer() -> None:
+                    await sleep(self.CPR_TIMEOUT)
 
                     # Not set in the meantime -> not supported.
                     if self.cpr_support == CPR_Support.UNKNOWN:
@@ -434,13 +451,11 @@ class Renderer(object):
 
                         if self.cpr_not_supported_callback:
                             # Make sure to call this callback in the main thread.
-                            get_event_loop().call_from_executor(self.cpr_not_supported_callback)
+                            self.cpr_not_supported_callback()
 
-                t = threading.Thread(target=timer)
-                t.daemon = True
-                t.start()
+                get_app().create_background_task(timer())
 
-    def report_absolute_cursor_row(self, row):
+    def report_absolute_cursor_row(self, row: int) -> None:
         """
         To be called when we know the absolute cursor position.
         (As an answer of a "Cursor Position Request" response.)
@@ -464,14 +479,14 @@ class Renderer(object):
             f.set_result(None)
 
     @property
-    def waiting_for_cpr(self):
+    def waiting_for_cpr(self) -> bool:
         """
         Waiting for CPR flag. True when we send the request, but didn't got a
         response.
         """
         return bool(self._waiting_for_cpr_futures)
 
-    def wait_for_cpr_responses(self, timeout=1):
+    async def wait_for_cpr_responses(self, timeout: int = 1) -> None:
         """
         Wait for a CPR response.
         """
@@ -479,34 +494,26 @@ class Renderer(object):
 
         # When there are no CPRs in the queue. Don't do anything.
         if not cpr_futures or self.cpr_support == CPR_Support.NOT_SUPPORTED:
-            return Future.succeed(None)
+            return None
 
-        f = Future()
-
-        # When a CPR has been received, set the result.
-        def wait_for_responses():
+        async def wait_for_responses() -> None:
             for response_f in cpr_futures:
-                yield From(response_f)
-            if not f.done():
-                f.set_result(None)
-        ensure_future(wait_for_responses())
+                await response_f
 
-        # Timeout.
-        def wait_for_timeout():
-            time.sleep(timeout)
+        async def wait_for_timeout() -> None:
+            await sleep(timeout)
 
-            # Got timeout.
-            if not f.done():
-                self._waiting_for_cpr_futures = deque()
-                f.set_result(None)
+            # Got timeout, erase queue.
+            self._waiting_for_cpr_futures = deque()
 
-        t = threading.Thread(target=wait_for_timeout)
-        t.daemon = True
-        t.start()
+        coroutines = [
+            wait_for_responses(),
+            wait_for_timeout(),
+        ]
+        await wait(coroutines, return_when=FIRST_COMPLETED)
 
-        return f
-
-    def render(self, app, layout, is_done=False):
+    def render(self, app: 'Application[Any]', layout: 'Layout',
+               is_done: bool = False) -> None:
         """
         Render the current interface to the output.
 
@@ -610,7 +617,7 @@ class Renderer(object):
         if is_done:
             self.reset()
 
-    def erase(self, leave_alternate_screen=True):
+    def erase(self, leave_alternate_screen: bool = True) -> None:
         """
         Hide all output and put the cursor back at the first line. This is for
         instance used for running a system command (while hiding the CLI) and
@@ -630,7 +637,7 @@ class Renderer(object):
 
         self.reset(leave_alternate_screen=leave_alternate_screen)
 
-    def clear(self):
+    def clear(self) -> None:
         """
         Clear screen and go to 0,0
         """
@@ -648,16 +655,14 @@ class Renderer(object):
 
 
 def print_formatted_text(
-        output, formatted_text, style, style_transformation=None,
-        color_depth=None):
+        output: Output,
+        formatted_text: AnyFormattedText,
+        style: BaseStyle,
+        style_transformation: Optional[StyleTransformation] = None,
+        color_depth: Optional[ColorDepth] = None) -> None:
     """
     Print a list of (style_str, text) tuples in the given style to the output.
     """
-    assert isinstance(output, Output)
-    assert isinstance(style, BaseStyle)
-    assert style_transformation is None or isinstance(style_transformation, StyleTransformation)
-    assert color_depth is None or color_depth in ColorDepth._ALL
-
     fragments = to_formatted_text(formatted_text)
     style_transformation = style_transformation or DummyStyleTransformation()
     color_depth = color_depth or ColorDepth.default()
@@ -671,7 +676,7 @@ def print_formatted_text(
         style.get_attrs_for_style_str,
         style_transformation)
 
-    for style_str, text in fragments:
+    for style_str, text, *_ in fragments:
         attrs = attrs_for_style_string[style_str]
 
         if attrs:
